from typing import Iterable


class HideClassInfo(type):
    __name__ = dict.__name__

    def __getattribute__(cls, name: str):
        if name == '__class__':
            return dict.__class__
        # Masquerading as a built-in thing will not serialize properly.
        if name == '__module__':
            return dict.__module__
        return super().__getattribute__(name)

    def __str__(cls) -> str:
        return str(dict)


class IsddcDict(dict, metaclass=HideClassInfo):

    def __init__(self, _data=None, **kw):
        if _data:
            _data.update(kw)
        else:
            _data = kw

        for name, value in _data.items():
            self[name] = IsddcDict(value)

    def __new__(cls, _data={}, **kw):
        if isinstance(_data, dict):
            return dict.__new__(cls)

        if isinstance(_data, Iterable) and not isinstance(_data, str):
            return [IsddcDict(v) for v in _data]
            # return _data.__class__(IsddcDict(v) for v in _data)

        return _data

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return self.__getitem__(name)

    # def __getattribute__(self, name):
    #     """
    #     `self.name` to `self[name]`
    #     So let's find the Key, and then let's find the built-in method.
    #     """
    #     try:
    #         return super().__getitem__(name)
    #     except KeyError:
    #         try:
    #             return super().__getattribute__(name)
    #         except AttributeError:
    #             raise KeyError(name)

    def __setattr__(self, name, value):
        self.__setitem__(name, value)

    def __delattr__(self, name):
        self.__delitem__(name)

    def __deepcopy__(self, memo):
        """
        This method must be implemented,  otherwise
        `KeyError: '__deepcopy__'` will appear when
        `copy.deepcopy (IsddcDict obj)` is called.
        """
        return IsddcDict(self)

    def __getstate__(self):
        """This method is for compatibility pickle module."""
        return self

    def __setstate__(self, state):
        """This method is for compatibility pickle module."""
        pass

    def __hash__(self):
        return -1

    def __eq__(self, other):
        return super().__eq__(other)

    def update(self, _data: dict, **kw):
        super().update(IsddcDict(_data, **kw))

    # This property is for compatibility abc.ABCMeta.
    __isabstractmethod__ = False

    # def __missing__(self, key):
    #     raise KeyError(f'{key}, {list(self)}')
